/*
 * CRRCsim - the Charles River Radio Control Club Flight Simulator Project
 *
 * Copyright (C) 2005-2010 Jan Reucker (original author)
 * Copyright (C) 2006, 2008 Jens Wilhelm Wulf
 * Copyright (C) 2008, 2009 Joel Lienard
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */
  

// implementation of class CGUILocationDialog

#include "../global.h"
#include "../SimStateHandler.h"
#include "../mod_landscape/crrc_scenery.h"
#include "crrc_gui_main.h"
#include "crrc_location.h"
#include "../crrc_main.h"
#include "../mod_mode/F3F/handlerF3F.h"
#include "../mod_windfield/windfield.h"
#include "../mod_misc/filesystools.h"
#include "util.h"
#include "../global_video.h"

#include <iostream>
#include <sstream>
#include <sys/types.h>
#include <dirent.h>
#include <string>


#define LIST_WIDGET_HEIGHT  (200)
#define LIST_WIDGET_WIDTH   (200)
#define BUTTON_BOX_HEIGHT   (2*DLG_DEF_SPACE+DLG_DEF_BUTTON_HEIGHT)
#define CONF_SEL_WIDTH      (200)
#define DESCRIPTION_HEIGHT  (75)

static void CGUILocationCallback(puObject *obj);
static void CGUILocationSelectCallback(puObject *obj);
//static void CGUISkySelectCallback(puObject *obj);

CGUILocationDialog::CGUILocationDialog() 
            : CRRCDialog(), cbox(NULL)
{
  // gather information on all installed sceneries
  DIR *dir=NULL;
  struct dirent *ent;
  const char* curLocName = Global::scenery->getName();
  int curSkyVariant = Global::scenery->getSkyVariant();
  int curLocIndex = -1;
  int lists_index = 0;
  std::vector<std::string> paths;
  std::vector<std::string> extlist;
  std::vector<std::string> fileslist;
  std::vector<std::string> locationslist;
  extlist.push_back("xml");
  
  T_Config::getLocationDirs(paths);
  
  for (unsigned int i = 0; i < paths.size(); i++)
  {
    if ((dir = opendir(paths[i].c_str())) == NULL)
    {
      #ifdef DEBUG_GUI
      std::cerr << "createFileList(): unable to open directory " << paths[i];
      std::cerr << std::endl; 
      #endif
    }
    else
    {
     while ((ent = readdir(dir)) != NULL)
      {
        std::string tmp;
        bool        fMatch = false;
        
        tmp = ent->d_name;
        
        for (unsigned int n=0; n<extlist.size() && fMatch == false; n++)
        {
          if (T_GUI_Util::checkExtension(tmp, extlist[n]))
            fMatch = true;
        }
                   
        if (fMatch)
        {
          std::string fullpath = paths[i] + "/" + tmp;
          //std::cout << fullpath << std::endl;
          SimpleXMLTransfer *loc = NULL;
          bool ok=false;
          std::string name;
          try
          {
            loc = new SimpleXMLTransfer(fullpath);
            name = loc->getChild("name")->getContentString();
            ok=true;
          }
          catch (XMLException e)
          {
            std::cerr << "Caught XML exception in CGUIlocationSelectDialog:" << std::endl;
            std::cerr << "  " << e.what() << std::endl;
            std::cerr << "  File: " << fullpath << std::endl;
          }
          if(ok)
          {
            fileslist.push_back(fullpath);
            locationslist.push_back(name);
            if (name == curLocName)
            {
              curLocIndex = lists_index;
            }
            lists_index++;
          }
          delete loc;
        }
      }
      closedir(dir);
    }
  }
  filesList = T_GUI_Util::loadnames(fileslist, filesListSize);
  locationsList = T_GUI_Util::loadnames(locationslist, locationsListSize);
  skiesList = NULL;
  skiesListSize = 0;

  // Now setup the GUI widgets
  // height of a text label
  int msg_height  = puGetDefaultLegendFont().getStringHeight("jX")
                  + puGetDefaultLegendFont().getStringDescender()
                  + PUSTR_TGAP + PUSTR_BGAP;
  
  // top of the list box
  int top_of_listbox = BUTTON_BOX_HEIGHT 
                        + 2*DLG_DEF_SPACE 
                        + msg_height
                        + LIST_WIDGET_HEIGHT
                        + DESCRIPTION_HEIGHT;

  // Scrolled ListBox for the scenery files
  cbox = new puaScrListBox (DLG_DEF_SPACE,
                            top_of_listbox - LIST_WIDGET_HEIGHT,
                            LIST_WIDGET_WIDTH,
                            LIST_WIDGET_HEIGHT, 
                            locationsList);

  cbox->setLabelPlace(PUPLACE_TOP_LEFT);
  cbox->setLabel("Select location:");
  cbox->setCallback(CGUILocationSelectCallback);
  cbox->setUserData(this);
  cbox->setValue(curLocIndex);

  ptext = new puText(DLG_DEF_SPACE, 
                     BUTTON_BOX_HEIGHT + DLG_DEF_SPACE);
  ptext->setLabel("File: /path/to/my/scenery.xml");

  
  // Combo box for sky variant selection
  sbox = new puaComboBox(3*DLG_DEF_SPACE + LIST_WIDGET_WIDTH,
                        top_of_listbox - DLG_DEF_BUTTON_HEIGHT,
                        3*DLG_DEF_SPACE + LIST_WIDGET_WIDTH + CONF_SEL_WIDTH,
                        top_of_listbox,
                        NULL,
                        false);
  sbox->setLabelPlace(PUPLACE_TOP_LEFT);
  sbox->setLabel("Select sky graphics:");
  sbox->setUserData(this);
  //sbox->setCallback(CGUISkySelectCallback);

  // the description box
  description = new puaLargeInput( DLG_DEF_SPACE,
                                  BUTTON_BOX_HEIGHT + DLG_DEF_SPACE,
                                  2*DLG_DEF_SPACE + LIST_WIDGET_WIDTH + CONF_SEL_WIDTH,
                                  DESCRIPTION_HEIGHT,
                                  1,    // num of arrow pairs
                                  16,   // slider width
                                  0);   // wrap text
  description->disableInput();
  description->setText("This is a short description of the selected scenery.");

  // finish the dialog
  close();
  setSize(LIST_WIDGET_WIDTH + CONF_SEL_WIDTH + 4*DLG_DEF_SPACE, 
          BUTTON_BOX_HEIGHT + LIST_WIDGET_HEIGHT + msg_height + 5*DLG_DEF_SPACE + DESCRIPTION_HEIGHT );
  setUserData(this);
  setCallback(CGUILocationCallback);

  // center the dialog on screen
  int wwidth, wheight;
  int current_width = getABox()->max[0] - getABox()->min[0];
  int current_height = getABox()->max[1] - getABox()->min[1];
  puGetWindowSize(&wwidth, &wheight);
  setPosition(wwidth/2 - current_width/2, wheight*2/3 - current_height);

  updateLocationInfo();
  sbox->setCurrentItem(curSkyVariant);

  reveal();
}

/**
 * Destroy the dialog.
 */
CGUILocationDialog::~CGUILocationDialog()
{
  T_GUI_Util::freenames(locationsList, locationsListSize);
  T_GUI_Util::freenames(filesList, filesListSize);
  T_GUI_Util::freenames(skiesList, skiesListSize);

}

/**
 * Update all widgets that depend on the currently selected location.
 */
void CGUILocationDialog::updateLocationInfo()
{
  std::vector<std::string> skynames;
  std::string filename = filesList[getLocationId()];
  SimpleXMLTransfer *xml = new SimpleXMLTransfer(filename);
  path_string = "File: ";
  path_string += filename;
  ptext->setLabel(path_string.c_str());
  int children = xml->getChildCount();
  std::vector<int> sky_indices;
  for (int i = 0; i < children; i++)
  {
    SimpleXMLTransfer *child = xml->getChildAt(i);
    if (child->getName() == "sky")
    {
      sky_indices.push_back(i);
      std::string desc = child->getChild("descr_short.en", true)->getContentString();
      if (desc == "")
      {
        std::ostringstream s;
        s << sky_indices.size() << ": no description";
        desc = s.str();
      }
      skynames.push_back(desc);
      //std::cout << "  <sky> " << sky_indices.size() << " at child idx " << i << ": " << desc<< std::endl;
    }
  }

  // Update the description box and launch default checkbox
  description_string = "";
  try
  {
    description_string = xml->getChild("description.en")->getContentString();
  }
  catch (XMLException &e)
  {
    description_string = "No description available.";
  }
  description_string = T_GUI_Util::trimWhitespace(description_string);
  description_string = T_GUI_Util::breakLines(description_string, 54);
  description->setText(description_string.c_str());
  description->setTopLineInWindow(0);
  
  delete xml;

  // If the scenery offers more than one sky to choose from,
  // update the combo box. Else hide it.
  if (skynames.size() > 1)
  {
    T_GUI_Util::freenames(skiesList, skiesListSize);
    skiesList = T_GUI_Util::loadnames(skynames, skiesListSize);
    sbox->newList(skiesList);
    sbox->reveal();
  }
  else
  {
    sbox->hide();
  }
  sbox->setCurrentItem(0);
}


std::string CGUILocationDialog::getLocation() const
{
  std::string loc = cbox->getStringValue();
  return loc;
}

int CGUILocationDialog::getLocationId() const
{
  return cbox->getIntegerValue();
}


bool CGUILocationDialog::saveSelection() const
{
  const char* curLocName = Global::scenery->getName();
  int curSkyVariant = Global::scenery->getSkyVariant();
  int id = getLocationId();
  std::string filename = filesList[id];
  std::string newLocname = locationsList[id];
  int newSkyVariant = sbox->getCurrentItem();
  if (newSkyVariant < 0)
  {
    // If the combo box was empty, the current item was -1.
    // In this case, correct the sky variant index to 0.
    newSkyVariant = 0;
  }


  if ((newLocname != curLocName) || (newSkyVariant != curSkyVariant))
  {
    cfg->setLocation(filename.c_str(), newSkyVariant, cfgfile);
      
    Scenery* new_scenery = loadScenery(FileSysTools::getDataPath(filename).c_str(),
                                       newSkyVariant);
    if (new_scenery)
    {
      clear_wind_field();
      delete Global::scenery;
      Global::scenery = new_scenery;
      //reinitialise game mode
      if (Global::gameHandler)
      {
        delete Global::gameHandler;
      }
      if (cfgfile->getInt("game.f3f.enabled",0))
        Global::gameHandler = new HandlerF3F();
      else
        Global::gameHandler= new T_GameHandler();
    }
    cfg->read(cfgfile);
    Video::setWindowTitleString();
    player_pos = Global::scenery->getPlayerPosition();
    Init_mod_windfield();
    Global::Simulation->reset();
  }
  return true;
}

/** \brief The dialog's callback.
 *
 *  Load the new location.
 */
void CGUILocationCallback(puObject *obj)
{
  if (obj->getIntegerValue() == CRRC_DIALOG_OK)
  {
    CGUILocationDialog *dlg = (CGUILocationDialog*)obj->getUserData();
    dlg->saveSelection();
  }
  puDeleteObject(obj);
  Global::gui->hide();
}

/** \brief The location list's callback.
 *
 *  Update scenery information whenever a new
 *  location is chosen.
 */
void CGUILocationSelectCallback(puObject *obj)
{
  CGUILocationDialog* dlg = static_cast<CGUILocationDialog*>(obj->getUserData());
  dlg->updateLocationInfo();
}

#if 0
void CGUISkySelectCallback(puObject *obj)
{
}
#endif
